import std.string;
import std.date;
import std.random;

import SDL;

import rc;
import log;
import fpstimer;
import scene;
import scene_game;
import scene_select;
import joystickmgr;
import padinput;
import colors;

class Cave {
private:
	static Cave obj_;
public:
	static Cave obj() {
		return obj_;
	}

public:
	this(char args[][]) {
		obj_ = this;
		Log.info("init");
		initResource(args);
		initMisc();
		initSDL();
		Log.info("init done");
	}

	~this() {
		Log.info("quit");
		delete joystickMgr_;
		SDL_Quit();
		Log.info("quit done");
	}

	void initResource(char args[][]) {
		Log.info("init resource");

		rc_ = new Rc();
		rc().addResourceFile();
		rc().addArgument(args);
	}

	void initMisc() {
		Log.info("init misc");

		rand_seed(getUTCtime(), 0);

		timer_ = new FpsTimer();
		timer_.setFps(60);
	}

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

		useAudio_ = true;
		ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
		if (ret != 0) {
			Log.warnSDL("cannot init audio");
			useAudio_ = false;
		}

		screen_ = SDL_SetVideoMode(rc().getInt("screenWidth"),
								   rc().getInt("screenHeight"),
								   16, SDL_SWSURFACE);
		if (screen_ == null) Log.errorSDL("cannot create screen");

		Colors.init(screen_);

		joystickMgr_ = new JoystickMgr();
		input_ = PadInput.generate("save/key.conf");

		background_ ~= MySDL_LoadBMP("data/bg1.bmp");
/*
		background_ ~= MySDL_LoadIMG("data/bg1.jpg");
		background_ ~= MySDL_LoadIMG("data/bg2.jpg");
		background_ ~= MySDL_LoadIMG("data/bg3.jpg");
		background_ ~= MySDL_LoadIMG("data/bg4.jpg");
		background_ ~= MySDL_LoadIMG("data/bg4.jpg");
		background_ ~= MySDL_LoadIMG("data/bg3.jpg");
		background_ ~= MySDL_LoadIMG("data/bg2.jpg");
		background_ ~= MySDL_LoadIMG("data/bg1.jpg");
*/
		for (int i = 0; i < background_.length; i++) {
			background_[i] = MySDL_DisplayFormat(background_[i]);
			SDL_SetAlpha(background_[i], SDL_SRCALPHA|SDL_RLEACCEL, 128);
		}
	}

	void run() {
		Log.info("run");

		scene_ = new SceneSelect();

		timer_.start();

		while (scene_ != null) {
			scene_.run();
			if (scene_.isEnd()) {
				scene_ = scene_.getNext();
				timer_.start();
			}

            SDL_UpdateRect(screen_, 0,0,0,0);
            int t = turn & 127;
            if ((t & 3) == 3) {
                //MySDL_BlitSurface(background_[t>>4], screen_);
                MySDL_BlitSurface(background_[0], screen_);
            }

			if (!(turn % 62)) {
				Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
			}
			timer_.wait();

			endCheck();
		}

		Log.info("run done");
	}

	void endCheck() {
		SDL_Event ev;
		while (SDL_PollEvent(&ev)) {
			if (ev.type == SDL_QUIT || input_.getButton(2)) {
				Log.info("quit event catched");
				quit();
			}
		}
	}

	void quit() {
		Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
		scene_ = null;
	}

	void end() {
		Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
		scene_ = scene_.getNext();
	}

public:
	static Rc rc() { return obj().rc_; }

	bit useAudio() { return useAudio_; }

	SDL_Surface* screen() { return screen_; }

	PadInput padInput() { return input_; }

    int turn() { return timer_.getTurn(); }

private:
	Rc rc_;

	FpsTimer timer_;

	JoystickMgr joystickMgr_;
	PadInput input_;

	bit useAudio_;

	SDL_Surface* screen_;

	Scene scene_;

	SDL_Surface*[] background_;
}

extern (C) int doit(char[][] args) {
	auto Cave c = new Cave(args);
	c.run();

	return 0;
}

version (linux) {
    int main(char[][] args) {
        return doit(args);
    }
}
