import std.string;
import std.conv;

import SDL;

import axis;
import log;
import joystickmgr;
import rc;

/// ܥϴ쥯饹
class ButtonInput {
public:
	bit getButton() { return false; }
	char[] getName() { return "error"; }

public:
	static ButtonInput generate(char[] str) {
		int unsco = find(str, cast(char)'_') + 1;
		if (unsco == -1) Log.error(str ~ ": invalid button key");

		char[] prefix = str[0..unsco];
		char[] val = str[unsco..str.length];

		SDL_Joystick[2] joys;
		if (prefix == "KEY_") {
			SDLKey key;

			if (val.length != 1) {
				if (val == "tab") key = SDLK_TAB;
				else if (val == "space") key = SDLK_SPACE;
				else if (val == "delete") key = SDLK_DELETE;
				else if (val == "return") key = SDLK_RETURN;
				else if (val == "escape") key = SDLK_ESCAPE;
				else if (val == "up") key = SDLK_UP;
				else if (val == "down") key = SDLK_DOWN;
				else if (val == "left") key = SDLK_LEFT;
				else if (val == "right") key = SDLK_RIGHT;
				else if (val == "insert") key = SDLK_INSERT;
				else if (val == "rshift") key = SDLK_RSHIFT;
				else if (val == "ralt") key = SDLK_RALT;
				else if (val == "rctrl") key = SDLK_RCTRL;
				else if (val == "lshift") key = SDLK_LSHIFT;
				else if (val == "lalt") key = SDLK_LALT;
				else if (val == "lctrl") key = SDLK_LCTRL;
				else Log.error(val ~ ": unknown key");
			}
			else {
				key = val[0];
			}

			return new KeyButtonInput(key);
		}
		else if (prefix == "JOY_" || prefix == "JOY1_" || prefix == "JOY2_") {
			int joynum = (prefix == "JOY2_") ? 1 : 0;
			if (JoystickMgr.obj().getJoystick(joynum)) {
				if (val.length != 1) {
					Axis axis;
					if (val == "UP") axis = new Axis(Axis.UP);
					else if (val == "RIGHT") axis = new Axis(Axis.RIGHT);
					else if (val == "DOWN") axis = new Axis(Axis.DOWN);
					else if (val == "LEFT") axis = new Axis(Axis.LEFT);
					else Log.error(val ~ ": unknown key");

					return new JoystickAxisButtonInput(joynum, axis);
				}
				else {
					int index = toInt(val);
					return new JoystickButtonInput(joynum, index);
				}
			}
			else {
				return null;
			}
		}
		else {
			Log.error("generateButtonInputFromString: unknown string format");
		}

		return null;
	}

};

/// ܡɤˤܥϥ饹
class KeyButtonInput : public ButtonInput {
public:
	this(SDLKey key) {
		key_ = key;
	}
	bit getButton() {
		return SDL_GetKeyState(null)[key_] != 0;
	}

	char[] getName() {
        auto char[] sdlk = std.string.toString(SDL_GetKeyName(key_));
		std.string.replace(sdlk, "right ", "");
		std.string.replace(sdlk, "left ", "");
        return "KEY_" ~ sdlk;
	}

private:
	SDLKey key_;
};

/// 祤ƥåˤܥϥ饹
class JoystickButtonInput : public ButtonInput {
public:
	this(int joystick, int button) {
		joystick_ = JoystickMgr.obj().getJoystick(joystick);
		button_ = button;
	}
	bit getButton() {
		return cast(bit)joystick_.getButton(button_);
	}

	char[] getName() {
		return "JOY_" ~ std.string.toString(button_);
	}

private:
	Joystick joystick_;
	int button_;
};

interface AxisInput {
public:
	Axis getAxis();
};

class JoystickAxisButtonInput : public ButtonInput {
public:
	static const int JOYSTICK_PLAY_WIDTH = 8192;

public:
	/**
	 * @param axis ФꤷƤϤʤʤ
	 */
	this(int index, Axis axis) {
		joystick_ = JoystickMgr.obj().getJoystick(index);
		axis_ = axis;
	}

	bit getButton() {
		if (axis_.isRight()) {
			if (joystick_.axisX() > JOYSTICK_PLAY_WIDTH) return true;
		}
		else if (axis_.isLeft()) {
			if (joystick_.axisX() < -JOYSTICK_PLAY_WIDTH) return true;
		}
		else if (axis_.isUp()) {
			if (joystick_.axisY() < -JOYSTICK_PLAY_WIDTH) return true;
		}
		else if (axis_.isDown()) {
			if (joystick_.axisY() > JOYSTICK_PLAY_WIDTH) return true;
		}

		return false;
	}

	char[] getName() {
		return "JOY_" ~ axis_.getName();
	}

private:
	Joystick joystick_;
	Axis axis_;
};

/// ButtonInput ˤϥ饹
class ButtonAxisInput : public AxisInput {
public:
	this() {}
	this(ButtonInput up, ButtonInput right,
		 ButtonInput down, ButtonInput left) {
		up_ ~= up;
		right_ ~= right;
		down_ ~= down;
		left_ ~= left;
	}

	Axis getAxis() {
		Axis ret = new Axis();

		for (int i = 0; i < up_.length; i++) {
			if (up_[i].getButton()) {
				ret.add(Axis.UP);
				break;
			}
		}

		for (int i = 0; i < down_.length; i++) {
			if (down_[i].getButton()) {
				ret.add(Axis.DOWN);
				break;
			}
		}

		for (int i = 0; i < right_.length; i++) {
			if (right_[i].getButton()) {
				ret.add(Axis.RIGHT);
				break;
			}
		}

		for (int i = 0; i < left_.length; i++) {
			if (left_[i].getButton()) {
				ret.add(Axis.LEFT);
				break;
			}
		}

		return ret;
	}

	void setUpAxis(ButtonInput button) {
		up_ ~= button;
	}
	void setRightAxis(ButtonInput button) {
		right_ ~= button;
	}
	void setDownAxis(ButtonInput button) {
		down_ ~= button;
	}
	void setLeftAxis(ButtonInput button) {
		left_ ~= button;
	}

private:
	ButtonInput[] up_, right_, down_, left_;
};

/// ϥ饹
class PadInput {
public:
	void setAxis(AxisInput axis) {
		axis_ ~= axis;
	}
	Axis getAxis() {
		Axis ret = new Axis();
		for (int i = 0; i < axis_.length; i++) {
			ret.add(axis_[i].getAxis());
		}

		return ret;
	}

	void setButton(int id, ButtonInput button) {
		button_[id] ~= button;
	}
	bit getButton(int id) {
		foreach (ButtonInput bi; button_[id]) {
			if (bi.getButton()) return true;
		}
		return false;
	}

private:
	AxisInput[] axis_;
	ButtonInput[][4] button_;

private:
	static PadInput generate(char[] filename) {
		Log.info("generate padinput from " ~ filename);

        PadInput ret = new PadInput;
		ButtonAxisInput axisInput = new ButtonAxisInput();

		auto Rc rc = new Rc;
        rc.addResourceFile(filename);

		char[][] confs = rc.keys();
		for (int i = 0; i < confs.length; i++) {
			char[] key = confs[i];
			char[][] vals = split(rc.get(key), " ");

			for (int j = 0; j < vals.length; j++) {
				char[] val = vals[j];

				ButtonInput b = ButtonInput.generate(val);

				if (!b) continue;

				if (key == "UP_AXIS") {
					axisInput.setUpAxis(b);
				}
				else if (key == "RIGHT_AXIS") {
					axisInput.setRightAxis(b);
				}
				else if (key == "DOWN_AXIS") {
					axisInput.setDownAxis(b);
				}
				else if (key == "LEFT_AXIS") {
					axisInput.setLeftAxis(b);
				}
				else if (key == "BUTTON_1") {
					ret.setButton(0, b);
				}
				else if (key == "BUTTON_2") {
					ret.setButton(1, b);
				}
				else if (key == "BUTTON_3") {
					ret.setButton(2, b);
				}
				else if (key == "END") {
					ret.setButton(3, b);
				}
			}
		}

		ret.setAxis(axisInput);
        return ret;

	}

};

