import std.string;
import std.math;

import cave;
import padinput;
import mymath;
import mysdl;
import log;
import game;
import charactor;
import command;
import mygfx;
import colors;

class Player : public Charactor {
public:
	this(Game game, PadInput input, int color) {
		game_ = game;
		input_ = input;

		color_ = color;
		opposite_ = Colors.opposite(color_);

		wait_ = 0;

		damageTurn_ = 0;

		char[] colStr = std.string.toString(color);
		x_ = Cave.rc().getFloat("playerX" ~ colStr);
		y_ = Cave.rc().getFloat("playerY" ~ colStr);

		playerSpeed_ = Cave.rc().getFloat("playerSpeed");
		playerSpeed2_ = Cave.rc().getFloat("playerSpeed2");

		playerSize_ = Cave.rc().getInt("playerSize");
		playerFieldSize_ = Cave.rc().getInt("playerFieldSize");
		playerFieldSize2_ = (playerFieldSize_ * playerFieldSize_) << 2;

		screenWidth_ = Cave.rc().getInt("screenWidth") - playerSize_;
		screenHeight_ = Cave.rc().getInt("screenHeight") - playerSize_;

		fieldSurf_ = MySDL_CreateRGBSurface(playerFieldSize_*2+1,
										   playerFieldSize_*2+1);
		filledCircleColor(fieldSurf_, playerFieldSize_, playerFieldSize_,
						  playerFieldSize_,
						  Colors.playerField[color]);
		SDL_SetColorKey(fieldSurf_, SDL_SRCCOLORKEY|SDL_RLEACCEL,
						SDL_MapRGB(fieldSurf_.format, 0,0,0));
		SDL_SetAlpha(fieldSurf_, SDL_SRCALPHA|SDL_RLEACCEL, 128);
		coreSurf_ = MySDL_CreateRGBSurface(playerSize_*2+1,
											playerSize_*2+1);
		filledCircleColor(coreSurf_, playerSize_, playerSize_, playerSize_,
						  Colors.playerCore[color]);
		SDL_SetColorKey(coreSurf_, SDL_SRCCOLORKEY|SDL_RLEACCEL,
						SDL_MapRGB(coreSurf_.format, 0,0,0));
		SDL_SetAlpha(coreSurf_, SDL_SRCALPHA|SDL_RLEACCEL, 128);
		gaugeSurf_ = MySDL_CreateRGBSurface(200, 20);
		SDL_FillRect(gaugeSurf_, cast(SDL_Rect*)0,
					 MyGFX_MapRGBA(gaugeSurf_, Colors.playerGauge[color]));
		SDL_SetAlpha(gaugeSurf_, SDL_SRCALPHA|SDL_RLEACCEL, 128);

		static char[][9] commandNames = [
			"neutral", "up", "upright", "right", "downright",
			"down", "downleft", "left", "upleft"
			];
        static int[9] whiteConv = [ 0, 5, 4, 3, 2, 1, 8, 7, 6 ];
		name_ = Cave.rc().get("playerCharactor" ~ Colors.colorStr(color_));
		auto LuaFile luaFile =
			new LuaFile("data/players/" ~ name_ ~ "/action.lua");
		for (int i = 0; i < commandNames.length; i++) {
            int j = i;
            if (color_ == Colors.WHITE && j != 0) j = whiteConv[j];
			main_[i] = new Command(luaFile,
								   commandNames[j]~"A", "mainA",
								   game_, this, color_);
			sub_[i] = new Command(luaFile,
								  commandNames[j]~"B", "mainB",
								  game_, this, color_);
		}

		gauge_ = 100;

	}

	void move() {
		turn_++;
		damageTurn_--;

		float base;

		if (input_ == null) return;

		bit button = input_.getButton(0);
		Axis axis = input_.getAxis();

		if (button) base = playerSpeed2_;
		else base = playerSpeed_;
		float cross = base / SQRT_2;

		if (axis == Axis.UP) y_ -= base;
		else if (axis == Axis.RIGHT) x_ += base;
		else if (axis == Axis.DOWN) y_ += base;
		else if (axis == Axis.LEFT) x_ -= base;
		else if (axis == Axis.UPRIGHT) {
			x_ += cross;
			y_ -= cross;
		}
		else if (axis == Axis.DOWNRIGHT) {
			x_ += cross;
			y_ += cross;
		}
		else if (axis == Axis.DOWNLEFT) {
			x_ -= cross;
			y_ += cross;
		}
		else if (axis == Axis.UPLEFT) {
			x_ -= cross;
			y_ -= cross;
		}

		if (axis != Axis.NONE) {
			Player opp = game_.player(opposite_);
			int d2 = cast(int)distance2(x_, y_, opp.x(), opp.y());
			if (d2 < playerFieldSize2_) {
				float dr = 1.01*sqrt((cast(float)playerFieldSize2_) / d2);
				int dx = cast(int)(x_ - opp.x());
				int dy = cast(int)(y_ - opp.y());
				x_ = opp.x() + dx * dr;
				y_ = opp.y() + dy * dr;
			}
		}

		if (x_ < playerSize_) x_ = playerSize_;
		else if (x_ > screenWidth_) x_ = screenWidth_;
		if (y_ < playerSize_) y_ = playerSize_;
		else if (y_ > screenHeight_) y_ = screenHeight_;

		if (wait_ == 0) {
			if (input_.getButton(0)) {
				main_[axis.getSmallAxisCode()].run();
			}
			if (input_.getButton(1)) {
				sub_[axis.getSmallAxisCode()].run();
			}
		}
		else {
			wait_--;
		}

		gauge_ += 1;
		if (gauge_ > 100) gauge_ = 100;
	}

	void drawField(SDL_Surface* scr, int color) {
		MySDL_BlitSurface(coreSurf_, scr,
                          cast(int)(x_-playerSize_), cast(int)(y_-playerSize_));
		if (damageTurn_ > 1) {
			static int flashCnt = 0;
			flashCnt = (flashCnt + 1) & 3;
			if (flashCnt) {
				return;
			}
		}
		MySDL_BlitSurface(fieldSurf_, scr,
						  cast(int)(x_-playerFieldSize_),
                          cast(int)(y_-playerFieldSize_));
		MySDL_BlitSurface(gaugeSurf_, 0, 0, cast(int)(gauge_*2), 20,
						  scr, 5, 615-color*610);
	}

	void setInput(PadInput input) {
		input_ = input;
	}

	void drawCore(SDL_Surface* scr, int color) {
		SDL_Rect rect = MySDL_Rect(cast(int)x_-2, cast(int)y_-2, 5, 5);
		SDL_FillRect(scr, &rect,
					 SDL_MapRGB(scr.format, 0, 255, 0));
	}

	float gauge() { return gauge_; }
	void useGauge(float g) { gauge_ -= g; }

	int lastBulletTurn() { return lastBulletTurn_; }
	void setLastBulletTurn(int t) { lastBulletTurn_ = t; }

	void wait(int t) {
		wait_ = t;
	}

	float playerSpeed() { return playerSpeed_; }
	float playerSpeed2() { return playerSpeed2_; }

	int screenWidth() { return screenWidth_; }
	int screenHeight() { return screenHeight_; }

	void getEnemyBullets(out int l, out float* x, out float* y,
						 out float* sx, out float* sy) {
		game_.getPlayerBullets(opposite_, l, x, y, sx, sy);
	}

	void damage() {
		if (damageTurn_ < 0) {
			Log.info(Colors.colorStr(color_) ~ ": damaged.");

			damageTurn_ = 60;
		}
	}

private:
	Game game_;
	PadInput input_;
	Command main_[9];
	Command sub_[9];

	float playerSpeed_;
	float playerSpeed2_;

	int color_;
	int opposite_;

	int playerSize_;
	int playerFieldSize_;
	int playerFieldSize2_;

	int screenWidth_;
	int screenHeight_;

	SDL_Surface* coreSurf_;
	SDL_Surface* fieldSurf_;
	SDL_Surface* gaugeSurf_;

	float gauge_;

	int lastBulletTurn_;

	int wait_;

	int damageTurn_;

	char[] name_;

}
